"""Helper functions for the pysma library."""

import dataclasses
import json
from typing import Any, Dict
from urllib.parse import urlparse


class BetterJSONEncoder(json.JSONEncoder):
    """JSON Encoder that handles dataclasses and non serialziable objects."""

    def default(self, o: Any) -> Any:
        """Handler for the Encoder."""
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return str(o)


def splitUrl(url: str, fallbackScheme: str = "fake") -> Dict[str, Any]:
    url = url.strip(" ")
    if "://" not in url:
        url = fallbackScheme + "://" + url
    x = urlparse(url)
    return {
        "schema": x.scheme,
        "host": x.hostname,
        "netloc": x.netloc,
        "port": x.port,
    }


def toJson(obj: Any):
    """Converts a object to a json String.
    Incl. handling of dataclass."""
    return json.dumps(obj, cls=BetterJSONEncoder, indent=4)


def version_int_to_string(version_integer: int) -> str:
    """Convert a version integer to a readable string.

    Args:
        version_integer (int): The version integer, as retrieved from the device.

    Returns:
        str: The version translated to a readable string.
    """
    if not version_integer:
        return ""

    appendixes = ["N", "E", "A", "B", "R", "S"]
    version_bytes = version_integer.to_bytes(4, "big")
    version_appendix = (
        appendixes[version_bytes[3]] if 0 <= version_bytes[3] < len(appendixes) else ""
    )
    return f"{version_bytes[0]:x}.{version_bytes[1]:x}.{version_bytes[2]}.{version_appendix}"


def isInteger(s: str) -> bool:
    """Test if the string is a integer"""
    try:
        int(s)
        return True
    except ValueError:
        return False



"""Exceptions for the pysma library."""


class SmaException(Exception):
    """Base exception of the pysma library."""


class SmaAuthenticationException(SmaException):
    """An attempt to authenticate failed."""


class SmaReadException(SmaException):
    """Reading the data did not return an expected result."""


class SmaWriteException(SmaException):
    """Writing the data did not return an expected result."""


class SmaConnectionException(SmaException):
    """An error occurred in the connection with the device."""



import collections
import copy
import html
import logging
from dataclasses import dataclass
from datetime import datetime
from typing import Awaitable, Callable

import untangle  # type: ignore
import xmlschema
from aiohttp import web

from pysmaplus.semp.const import callbackAction

from .const import (
    descriptionXML,
    deviceInfoXML,
    deviceStatusXML,
    sempXMLend,
    sempXMLstart,
    timeFrameXml,
)
from .device import sempDevice
from .RendererStatusPage import statusPageRenderer
from .sempxsd import sempxsd

_LOGGER = logging.getLogger(__name__)


@dataclass
class historyData:
    time: str
    timemsec: float
    typ: str
    remote: str
    deviceData: dict[str, sempDevice] | None
    cmdData: dict | None


class SEMPhttpServer:
    routes = web.RouteTableDef()
    devices: dict[str, sempDevice] = {}
    history: collections.deque[historyData] = collections.deque(maxlen=180)

    def __init__(
        self,
        addr: str,
        port: int,
        uuid: str,
        timezone,
        callback: Callable[[callbackAction], Awaitable[None]] | None = None,
    ):
        """Initialize the instance of the view."""
        print("Init HTTP-Server")
        self.port = port
        self.addr = addr
        self.lastip = addr.split(".")[-1]
        self.uuid = uuid
        self.timezone = timezone
        self.sempSchema: xmlschema.XMLSchema10 = xmlschema.XMLSchema(sempxsd)
        self.callback = callback

    async def getUUIDPage(self, request: web.Request) -> web.Response:

        msg = descriptionXML.format(
            friendly_name=self.lastip + " HA-GW",
            manufacturer="Home Assistant",
            manufacturer_url="https://www.home-assistant.io/",
            model_description="SEMP-to-Home-Assistant-Gateway",
            model_name="HA_SEMP",
            model_number="1.0.0",
            model_url="https://www.home-assistant.io/",
            serial_number="123",  # TODO
            uuid="53-4D-41-53-4D-43",  # TODO
            presentation_url=f"http://{self.addr}:{self.port}",
        )
        return web.Response(text=msg, content_type="text/xml")

    async def getStatusPage(self, request: web.Request):
        return await statusPageRenderer(request, self.devices, self.history)

    def bool2str(self, b) -> str:
        return str(b).lower()

    async def postSemp(self, request) -> web.Response:
        buffer = b""
        async for data, end_of_http_chunk in request.content.iter_chunks():
            buffer += data
            # if end_of_http_chunk:
            #     print(buffer)
        #        try:
        _LOGGER.info(buffer)
        print(buffer)
        #  '<EM2Device xmlns="http://www.sma.de/communication/schema/SEMP/v1"><DeviceControl>
        # <DeviceId>F-00000001-000000000002-00</DeviceId>
        # <On>true</On>
        # </DeviceControl></EM2Device>
        # b'<?xml version="1.0" encoding="UTF-8"?>\n<EM2Device xmlns="http://www.sma.de/communication/schema/SEMP/v1">\n\t<DeviceControl>\n\t\t<DeviceId>F-11223344-223529992949-00</DeviceId>\n\t\t<On>true</On>\n\t\t<RecommendedPowerConsumption>0</RecommendedPowerConsumption>\n\t\t<Timestamp>0</Timestamp>\n\t</DeviceControl>\n</EM2Device>'

        obj = untangle.parse(buffer.decode("utf-8"))
        assert (
            obj.children[0]["xmlns"] == "http://www.sma.de/communication/schema/SEMP/v1"
        ), "Not a SEMP XML-File received."
        devId = obj.EM2Device.DeviceControl.DeviceId.cdata
        onoff = obj.EM2Device.DeviceControl.On.cdata.lower()
        if onoff == "true":
            onoffStr = "on"
        else:
            onoffStr = "off"
        data = {}
        data[devId] = {}
        data[devId]["devid"] = devId
        data[devId]["status"] = onoffStr
        h = historyData(
            datetime.now().isoformat()[0:19],
            datetime.now().timestamp(),
            "COMMAND",
            str(request.remote),
            None,
            data,
        )
        self.history.append(h)
        assert onoff in ["true", "false"]
        onoffBool = onoff == "true"
        # print(devId, onoffBool)
        # if devId.startswith("F-11223344-") and devId.endswith("-00"):
        #     devId = devId[2 + 8 + 1 : -3]
        # else:
        #     _LOGGER.warning(f"Unknown device id received {devId}")
        if self.callback:
            shortDevId = devId[2 + 8 + 1 : -3]
            _LOGGER.info(f"Command received {onoffBool} {devId} {shortDevId}")
            await self.callback(callbackAction(devId, shortDevId, onoffBool))
        return web.Response(text="", content_type="text/xml")

    def Now(self) -> datetime:
        if self.timezone:
            return datetime.now(tz=self.timezone)
        else:
            return datetime.now()

    async def getSemp(self, request) -> web.Response:
        now = self.Now()

        msg = sempXMLstart
        data = {}
        for dev in self.devices.values():
            dev.update()
            data[dev.deviceId] = copy.copy(dev)
            msg += deviceInfoXML.format(
                deviceId=dev.deviceId,
                deviceName=html.escape(dev.deviceName),
                deviceType=dev.deviceType,
                deviceSerial=dev.deviceSerial,
                deviceVendor=html.escape(dev.deviceVendor),
                maxPowerConsumption=int(dev.deviceMaxConsumption),
                minPowerConsumption=int(dev.deviceMinConsumption),
                interruptionsAllowed=self.bool2str(dev.interruptionsAllowed),
                optionalEnergy=self.bool2str(dev.optionalEnergy),
                minOnTime=int(dev.minOnTime),
                minOffTime=int(dev.minOffTime),
            )
        now = self.Now()
        h = historyData(
            now.isoformat()[0:19],
            now.timestamp(),
            "STATUS",
            str(request.remote),
            data,
            None,
        )
        self.history.append(h)
        timeframeExists = False
        for dev in self.devices.values():
            msg += deviceStatusXML.format(
                power=dev.power,
                deviceName=html.escape(dev.deviceName),
                status=dev.status,
                deviceId=dev.deviceId,
                emSignalsAccepted=self.bool2str(dev.emSignalsAccepted),
            )
            if len(dev.timeframes) > 0:
                timeframeExists = True

        # Create PlanningRequest if at least one timeframe exists for all devices
        if timeframeExists:
            msg += "<PlanningRequest>"
            for dev in self.devices.values():
                for tf in dev.timeframes:
                    msg += timeFrameXml.format(
                        deviceId=dev.deviceId,
                        deviceName=dev.deviceName,
                        minrunningtime=tf.minRunningTime,
                        maxrunningtime=tf.maxRunningTime,
                        start=tf.start,
                        stop=tf.stop,
                        startSec=max(0, int((tf.start - now).total_seconds())),
                        stopSec=max(0, int((tf.stop - now).total_seconds())),
                    )
            msg += "</PlanningRequest>"
        msg += sempXMLend
        self.sempSchema.validate(msg)
        return web.Response(text=msg, content_type="text/xml")

    async def startWebserver(self) -> None:
        async def middleware_factory(app, handler):
            async def middleware_handler(request):
                print(request)
                print(
                    f"Request: {request.method} {request.rel_url} {request.remote} {request.raw_headers}"
                )
                return await handler(request)

            return middleware_handler

        #        logging.basicConfig(level=logging.ERROR)
        app = web.Application(middlewares=[middleware_factory])

        # TODO
        # Request all information (DeviceInfo, DeviceStatus, PlanningRequests of all devices):
        # o HTTP GET: <baseURL>/
        # Request DeviceInfo of all devices:
        # o HTTP GET: <baseURL>/DeviceInfo
        # Request DeviceInfo of specific device:
        # o HTTP GET: <baseURL>/DeviceInfo?DeviceId=<DeviceID>
        # Request DeviceStatus of all devices:
        # o HTTP GET: <baseURL>/DeviceStatus
        # Request DeviceStatus of specific device:
        # o HTTP GET: <baseURL>/DeviceStatus?DeviceId=<DeviceID>
        # Request PlanningRequest of all devices:
        # o HTTP GET: <baseURL>/PlanningRequest
        # Request PlanningRequest of specific device:
        # o HTTP GET: <baseURL>/PlanningRequest?DeviceId=<DeviceID>
        app.add_routes(
            [
                web.get("/", self.getStatusPage),
                web.get("/uuid:" + self.uuid + "/description.xml", self.getUUIDPage),
                web.get("/semp/", self.getSemp),
                web.post("/semp/", self.postSemp),
            ]
        )
        self.runner = web.AppRunner(app)
        await self.runner.setup()
        site = web.TCPSite(self.runner, self.addr, self.port)
        await site.start()

    def addDevice(self, device: sempDevice) -> None:
        self.devices[device.deviceId] = device

    def removeDevice(self, device: sempDevice) -> None:
        print(device)
        print(type(device))
        del self.devices[device.deviceId]

    def getDevice(self, deviceId: str) -> sempDevice | None:
        print(self.devices, deviceId)
        return self.devices.get(deviceId, None)


# sempXMLend



"""Interface for SMA ennoxOS based devices.

e.g. Tripower X and maybe EV Charger.
"""

import asyncio
import copy
import json
import logging
from dataclasses import asdict, dataclass, field
from datetime import UTC, datetime
from typing import Any, Dict, Optional

from aiohttp import ClientSession, ClientTimeout, client_exceptions, hdrs

from .const import SMATagList
from .definitions_ennexos import DEFAULT_TIMEOUT, getSensorForDevice
from .device import Device, DeviceInformation, DiscoveryInformation
from .exceptions import (
    SmaAuthenticationException,
    SmaConnectionException,
    SmaReadException,
)
from .helpers import splitUrl
from .sensor import Sensor, Sensor_Range, Sensors

_LOGGER = logging.getLogger(__name__)


@dataclass
class EnnexosDebug:
    parameters: dict[str, Any] = field(default_factory=dict)
    parameters_raw: dict[str, Any] = field(default_factory=dict)
    measurements: dict[str, Any] = field(default_factory=dict)
    measurements_raw: dict[str, Any] = field(default_factory=dict)
    last_notfound: dict[str, list] = field(default_factory=dict)
    profilemissing: dict[str, list] = field(default_factory=dict)
    plantInfo: dict[str, Any] = field(default_factory=dict)
    devInfo: dict[str, Any] = field(default_factory=dict)


class SMAennexos(Device):
    """Class to connect to the ennexos based SMA inverters."""

    # pylint: disable=too-many-instance-attributes
    _debug: EnnexosDebug = EnnexosDebug()

    _aio_session: ClientSession
    _new_session_data: Optional[dict[str, Any]]
    _url: str
    _authorization_header: dict[str, str]
    _componentId = "IGULD:SELF"
    _device_list: Dict[str, DeviceInformation] = {}

    def __init__(
        self,
        session: ClientSession,
        url: str,
        password: str | None,
        group: str | None,
    ):
        """Init SMA connection.

        Args:
            session (ClientSession): aiohttp client session
            url (str): Url or IP address of device
            password (str, optional): Password to use during login.
            group (str, optional): Username to use during login.

        """
        destination = splitUrl(url, "https")
        self._url = destination["schema"] + "://" + destination["host"]
        if destination["port"] is not None:
            self._url = self._url + ":" + str(destination["port"])
        _LOGGER.debug(f"Ennexos {url} => {self._url}")
        self._new_session_data = {"user": group, "pass": password}
        self._aio_session = session

    async def _jsonrequest(
        self, url: str, parameters: Dict[str, Any], method: str = hdrs.METH_POST
    ) -> Any:
        """Request json data for requests.

        Args:
            url (str): URL to do request to
            parameters (Dict[str, Any]): parameters
            method (str): Post or Get-Request

        Raises:
            SmaConnectionException: Connection to device failed
            SmaAuthenticationException: Authentication failed

        Returns:
            dict: json returned by device
        """
        try:
            async with self._aio_session.request(
                method, url, timeout=ClientTimeout(total=DEFAULT_TIMEOUT), **parameters
            ) as res:
                _LOGGER.debug(f"Request {url} Code {res.status}")
                if res.status == 200:
                    res = await res.json()
                    return res
                elif res.status == 401 or res.status == 400:
                    res = await res.json()
                    _LOGGER.error("Error " + str(res.status))
                    _LOGGER.error(res)
                    raise SmaAuthenticationException("Token failed!")
                else:
                    _LOGGER.warning("HTTP-Error %d for %s", res.status, url)
                    return {}
        except SmaAuthenticationException as e:
            raise e
        except (client_exceptions.ContentTypeError, json.decoder.JSONDecodeError):
            _LOGGER.error("Request to %s did not return a valid json.", url)
        except client_exceptions.ServerDisconnectedError as exc:
            _LOGGER.error(
                f"Timeout while requesting {url} {exc}. (Probably wrong ip, wrong hostname or wrong setting for http/https)"
            )
            raise SmaConnectionException(
                f"Server at {self._url} disconnected."
            ) from exc
        except (
            client_exceptions.ClientError,
            asyncio.exceptions.TimeoutError,
        ) as exc:
            _LOGGER.error(f"Error requesting {url} {exc} [Timeout]")
            raise SmaConnectionException(
                f"Could not connect to SMA at {self._url}: {exc}"
            ) from exc
        return {}

    # @override
    async def new_session(self) -> bool:
        """Establish a new session.

        Returns:
            bool: authentication successful
        """
        if self._new_session_data is None:
            _LOGGER.error("User & Pwd not set!")
            return False
        _LOGGER.debug(f'Trying to login {self._url} {self._new_session_data["user"]}')
        loginurl = self._url + "/api/v1/token"
        postdata = {
            "data": {
                "grant_type": "password",
                "username": self._new_session_data["user"],
                "password": self._new_session_data["pass"],
            }
        }
        ret = await self._jsonrequest(loginurl, postdata)
        if "access_token" not in ret:
            _LOGGER.debug(f"Login failed {ret}")
            raise SmaAuthenticationException("Login failed!")
        self._authorization_header = {
            "Authorization": "Bearer " + ret["access_token"],
            "Content-Type": "application/json",
        }
        _LOGGER.debug("Login successful")

        for u in [
            "/api/v1/plants/Plant:1",
            "/api/v1/plants/Plant:1/devices",
            "/api/v1/featuretoggles",
        ]:
            url = self._url + u
            try:
                ret = await self._jsonrequest(
                    url, {"headers": self._authorization_header}, hdrs.METH_GET
                )
            except SmaConnectionException as e:
                # Ignore Error while requesting featuretoogles as not available on all systems
                if u == "/api/v1/featuretoggles":
                    continue
                raise e
            if isinstance(ret, dict):
                ret.pop("_links", None)
            if isinstance(ret, list):
                for i in ret:
                    if isinstance(i, dict):
                        i.pop("_links", None)

            self._debug.plantInfo[u.split("/")[-1]] = ret

        return True

    async def _get_parameter(self, componentId: str) -> Dict[str, Dict[str, Any]]:
        """Get all parameters from the device.

        Returns:
            Dict: Return a dict with all parameters

        """
        liveurl = self._url + "/api/v1/parameters/search"
        postdata = {
            "data": '{"queryItems":[{"componentId":"' + componentId + '"}]}',
            "headers": self._authorization_header,
        }
        ret = await self._jsonrequest(liveurl, postdata)
        data = await self._prepare_parameter(ret, componentId)
        return data

    async def _prepare_parameter(
        self, ret: Any, componentId: str
    ) -> Dict[str, Dict[str, Any]]:
        data: Dict[str, Dict[str, Any]] = {}
        if len(ret) != 1:
            _LOGGER.warning(
                "Uncommon length of array in parameters request: %d", len(ret)
            )

        if len(ret) > 0:
            for d in ret[0]["values"]:
                dname = d["channelId"].replace("Parameter.", "").replace("[]", "")
                if "value" in d:
                    v = d["value"]
                    sensor_range = Sensor_Range("", [], d["editable"])
                    if "min" in d and "max" in d:
                        sensor_range = Sensor_Range(
                            "min/max", [d["min"], d["max"]], d["editable"]
                        )
                    if "possibleValues" in d:
                        sensor_range = Sensor_Range(
                            "selection", d["possibleValues"], d["editable"]
                        )

                    data[dname] = {
                        "name": dname,
                        "value": v,
                        "origname": d["channelId"],
                        "range": sensor_range,
                    }

                elif "values" in d:
                    # Split Value-Arrays
                    for idx in range(0, len(d["values"])):
                        v = d["values"][idx]
                        idxname = dname + "." + str(idx + 1)
                        data[idxname] = {
                            "name": idxname,
                            "value": v,
                            "origname": d["channelId"],
                        }
                else:
                    # Value current not available // night?
                    pass
        self._debug.parameters_raw[componentId] = ret
        self._debug.parameters[componentId] = data
        return data

    async def _get_all_readings(self, deviceID: str) -> Dict[str, Dict[str, Any]]:
        readings = await self._get_livedata(deviceID)
        readings.update(await self._get_parameter(deviceID))
        return readings

    async def _get_livedata(self, componentId: str) -> Dict[str, Dict[str, Any]]:
        """Get the sensors reading from the device.

        Returns:
            Dict: Return a dict with all measurements

        """
        liveurl = self._url + "/api/v1/measurements/live"
        postdata = {
            "data": '[{"componentId":"' + componentId + '"}]',
            "headers": self._authorization_header,
        }
        ret = await self._jsonrequest(liveurl, postdata)
        out = await self._prepare_livedata(ret, componentId)
        return out

    async def _prepare_livedata(
        self, ret: Any, componentId: str
    ) -> Dict[str, Dict[str, Any]]:
        """Convert the raw data from the inverter to a dict"""
        data: Dict[str, Any] = {}
        for d in ret:
            dname = d["channelId"].replace("Measurement.", "").replace("[]", "")
            if "value" in d["values"][0]:
                v = d["values"][0]["value"]
                if self._isfloat(v):
                    v = round(v, 2)
                data[dname] = {"name": dname, "value": v, "origname": d["channelId"]}
            elif "values" in d["values"][0]:
                # Split Value-Arrays
                for idx in range(0, len(d["values"][0]["values"])):
                    v = d["values"][0]["values"][idx]
                    if self._isfloat(v):
                        v = round(v, 2)
                    idxname = dname + "." + str(idx + 1)
                    data[idxname] = {
                        "name": idxname,
                        "value": v,
                        "origname": d["channelId"],
                    }
            else:
                # Value current not available // night?
                pass
        self._debug.measurements_raw[componentId] = ret
        self._debug.measurements[componentId] = data
        return data

    # @override
    async def get_sensors(self, deviceID: str | None = None) -> Sensors:
        """Get the sensors that are present on the device.

        Returns:
            Sensors: Sensors object containing Sensor objects
        """
        if not self._device_list:
            raise SmaReadException("device_info() not called!")
        deviceID = self.deviceIDFallback(deviceID)

        # Mark sure this function get called at least once
        ret = await self._get_all_readings(deviceID)
        _LOGGER.debug("Found Sensors for %s: %d", deviceID, len(ret))
        profile = await self._get_sensor_profile(deviceID)
        return profile

    async def _get_sensor_profile(self, deviceID: str) -> Sensors:
        device_sensors = Sensors()

        # Search for matiching profile
        dev = self._device_list[deviceID]
        productTagId = int(dev.additional.get("productTagId", 0))
        profile = getSensorForDevice(productTagId)
        if not profile:
            _LOGGER.warning(
                f"Unknown Device: {productTagId} N:{dev.name} T:{dev.type} ID:{deviceID}"
            )
            return device_sensors
        expected_sensors, unknown = profile
        if len(unknown) > 0:
            _LOGGER.debug(f"Missing Sensors in Profile {productTagId}: {unknown}")
            self._debug.profilemissing[deviceID] = unknown
        # Add Sensors from profile
        for s in expected_sensors:
            if s.name:
                device_sensors.add(copy.copy(s))
        return device_sensors

    async def close_session(self) -> None:
        """Closes the session."""

    def _isfloat(self, num: Any) -> bool:
        """Test if num is a float.

            Tests for type float or a string with a dot is is float

        Args:
            num: number to check

        Returns:
            bool: true, if num is from type float or a string with a dot
        """
        if isinstance(num, float):
            return True
        if isinstance(num, int):
            return False
        if not isinstance(num, str):
            raise TypeError("Value is not a string, float or int!")
        if "." not in num:
            return False
        try:
            float(num)
            return True
        except ValueError:
            return False

    def deviceIDFallback(self, deviceID: str | None) -> str:
        if not deviceID:
            return self._componentId
        return deviceID

    # @override
    async def read(self, sensors: Sensors, deviceID: str | None = None) -> bool:
        """Read a set of keys.

        Args:
            sensors (Sensors): Sensors object containing Sensor objects to read

        Returns:
            bool: reading was successful
        """
        notfound = []
        deviceID = self.deviceIDFallback(deviceID)
        data = None
        try:
            data = await self._get_all_readings(deviceID)
        except SmaAuthenticationException:
            # Relogin
            _LOGGER.debug("Re-login .. Starting new Session")
            await self.new_session()
            data = await self._get_all_readings(deviceID)
        for sen in sensors:
            if sen.enabled:
                if sen.key in data:
                    value = data[sen.key]["value"]
                    if sen.mapper:
                        sen.mapped_value = sen.mapper.get(value, str(value))
                    if sen.factor and sen.factor != 1:
                        value = round(value / sen.factor, 4)
                    sen.value = value
                    if "range" in data[sen.key]:
                        sen.range = data[sen.key]["range"]
                    continue
                notfound.append(f"{sen.name} [{sen.key}]")

        if deviceID not in self._debug.last_notfound:
            self._debug.last_notfound[deviceID] = []

        notfound = [x for x in notfound if x not in self._debug.last_notfound[deviceID]]
        if len(notfound) > 0:
            _LOGGER.info(
                "No values for sensors: %s",
                ",".join(notfound),
            )
            self._debug.last_notfound[deviceID].extend(notfound)
        return True

    # @override
    async def device_info(self) -> dict:
        """Read device info and return the results.

        Returns:
            dict: dict containing serial, name, type, manufacturer and sw_version
        """
        data = await self.device_list()
        if self._componentId in data:
            return data[self._componentId].asDict()
        else:
            return {}

    # @override
    async def device_list(self) -> dict[str, DeviceInformation]:

        devices = await self._jsonrequest(
            self._url + "/api/v1/plants/Plant:1/devices",
            {"headers": self._authorization_header},
            hdrs.METH_GET,
        )
        deviceID = {"IGULD:SELF"}
        for d in devices:
            deviceID.add(d["deviceId"])
        self._device_list = {}
        for d in deviceID:
            di = await self._device_info_by_componentId(d)
            if di:
                self._device_list[di.id] = di

        # TO DO move to device info by commpeontID TO DO
        # Adding Device-Placeholder for the whole plant
        pi = self._debug.plantInfo["Plant:1"]
        serial = "P" + self._device_list["IGULD:SELF"].serial
        self._device_list["Plant:1"] = DeviceInformation(
            "Plant:1", serial, pi["name"], pi["plantId"], "SMA", ""
        )
        self._device_list["Plant:1"].additional["productTagId"] = -47114711
        return self._device_list

    async def _device_info_by_componentId(
        self, componentId: str
    ) -> DeviceInformation | None:
        url = self._url + "/api/v1/plants/Plant:1/devices/" + componentId
        devInfo = await self._jsonrequest(
            url, {"headers": self._authorization_header}, hdrs.METH_GET
        )
        self._debug.devInfo[componentId] = devInfo
        if not devInfo:
            return None
        di = DeviceInformation(
            componentId,
            devInfo["serial"],
            devInfo["product"],
            devInfo["name"],
            devInfo["vendor"],
            devInfo["firmwareVersion"],
        )
        para = await self._get_parameter(componentId)
        di.parameterCount = len(para)
        data = await self._get_livedata(componentId)
        di.measurementsCount = len(data)
        for key, value in devInfo.items():
            if key in [
                "serial",
                "product",
                "name",
                "vendor",
                "firmwareVersion",
                "_links",
            ]:
                continue
            if type(value) not in [str, int, float]:
                continue
            di.additional[key] = value
        return di

    # @override
    async def get_debug(self) -> Dict:
        """Returns all Debug Information."""
        x = asdict(self._debug)
        x["device_list"] = self._device_list
        return x

    # @override
    async def detect(self, ip: str) -> list[DiscoveryInformation]:
        """Tries to detect a ennexos-based Device on this ip-address."""
        rets = []
        for prefix in ["https", "http"]:
            di = DiscoveryInformation()
            rets.append(di)
            url = f"{prefix}://{ip}/api/v1/system/info"
            di.tested_endpoints = url
            di.remark = prefix
            try:
                dev = await self._jsonrequest(url, {}, hdrs.METH_GET)
                if "productFriendlyNameTagId" in dev:
                    fallback = "Unknown: " + str(dev["productFriendlyNameTagId"])
                    di.device = SMATagList.get(
                        dev["productFriendlyNameTagId"], fallback
                    )
                    di.status = "found"
                    break
                di.status = "failed"
                di.exception = None
            except Exception as e:  # pylint: disable=broad-exception-caught
                di.status = "failed"
                di.exception = e
        return rets

    # @override
    def set_options(self, options: Dict[str, Any]) -> None:
        """Set low-level options."""
        for key, value in options.items():
            if key == "componentId":
                print(f"Option {key}: {self._componentId} => {value}")
                self._componentId = value
            else:
                _LOGGER.error("Unknown Options: %s %s", key, value)

    async def _get_timestamp(self) -> str:
        """Returns the time in a format as required by the put instruction."""
        return (
            f"{datetime.now(tz=UTC).isoformat(timespec='milliseconds').split('+')[0]}Z"
        )

    # @override
    async def set_parameter(
        self, sensor: Sensor, value: int, deviceID: str | None = None
    ) -> None:
        """SetParameters."""
        if deviceID not in self._device_list:
            raise RuntimeError("DeviceID %s unknown.", deviceID)
        timestamp = await self._get_timestamp()
        parameters = await self._get_parameter(deviceID)
        channelName = parameters[sensor.key]["origname"]
        requestData = f'{{"values":[{{"channelId":"{channelName}","timestamp":"{timestamp}","value":"{value}"}}]}}'
        putdata = {
            "data": requestData,
            "headers": self._authorization_header,
        }
        url = self._url + "/api/v1/parameters/" + deviceID
        dev = await self._jsonrequest(url, putdata, hdrs.METH_PUT)  # noqa: F841



"""Interface for SMA Energy Meters and Sunny Home Manager 2 (SHM2)

see https://www.unifox.at/software/sma-em-daemon/
see https://cdn.sma.de/fileadmin/content/www.developer.sma.de/docs/EMETER-Protokoll-TI-en-10.pdf

"""

import asyncio
import base64
import copy
import logging
import socket
import struct
import time
from dataclasses import dataclass, field
from typing import Any, Dict, List

from .const import SMATagList
from .definitions_em import obis2sensor
from .definitions_speedwire import speedwireHeader, speedwireHeader6069
from .device import Device, DeviceInformation, DiscoveryInformation
from .exceptions import SmaConnectionException, SmaReadException
from .sensor import Sensor, Sensors

_LOGGER = logging.getLogger(__name__)


@dataclass
class Debug_information_em:
    """Struct to store debug Information"""

    serial: set[int] = field(default_factory=set)
    protocol: set[str] = field(default_factory=lambda: set())
    last_packet_metadata: tuple[str, int, int] | None = None
    last_packet: bytes | None = None
    last_data: dict[str, Any] | None = None
    last_valid_packet: bytes | None = None


class SMAspeedwireEM(Device):
    """Class for the detection of SMA Devices in the local network."""

    def __init__(self) -> None:
        """init"""
        self.loop = asyncio.get_event_loop()
        self._transport: asyncio.BaseTransport | None = None
        self._protocol: SMAspeedwireEM | None = None
        self.transport: asyncio.BaseTransport | None = None
        self._data_received: asyncio.Future | None = None
        self.di = Debug_information_em()
        self._device_list: Dict[str, DeviceInformation] = {}
        self._expected_device: str | None = None
        self._bindingAddr: List[Any] = []

    # @override
    async def get_sensors(self, deviceID: str | None = None) -> Sensors:
        """Get the sensors that are present on the device.

        Returns:
            Sensors: Sensors object containing Sensor objects

        """
        device_sensors = Sensors()
        for s in obis2sensor:
            if s.name is not None:
                device_sensors.add(copy.copy(s))
        return device_sensors

    async def new_session(self) -> bool:
        """Starts a new session"""
        sock = self._getDiscoverySocket()
        on_connection_lost = self.loop.create_future()  # noqa F841
        self._transport, self._protocol = await self.loop.create_datagram_endpoint(
            lambda: self,  # type: ignore[type-var]
            sock=sock,
        )
        data = None
        try:
            data = await self._get_next_values()
        except TimeoutError as e:
            raise SmaConnectionException("No speedwire packet received!") from e
        if not data:
            raise SmaReadException("No usable data received!")
        return True

    async def _get_next_values(
        self, deviceID: str | None = None, timeout: float = 2
    ) -> dict:
        """Returns the next values received from the device."""
        self._data_received = asyncio.get_running_loop().create_future()
        self._expected_device = deviceID
        await asyncio.wait_for(self._data_received, timeout=timeout)
        data = self._data_received.result()
        self._data_received = None
        return data

    # @override
    async def device_info(self) -> dict:
        """Read device info and return the results.

        Returns:
            dict: dict containing serial, name, type, manufacturer and sw_version
        """
        #        await asyncio.wait_for(self._data_received, timeout=timeout)

        di = await self.device_list()
        return list(di.values())[0].asDict()

    # @override
    async def device_list(self) -> dict[str, DeviceInformation]:
        """Read device info and return the results.

        Returns:
            dict: dict containing serial, name, type, manufacturer and sw_version
        """
        self._device_list = {}
        start = time.time()
        while time.time() - start < 2.1:
            data = await self._get_next_values()
            if data["serial"] not in self._device_list:
                di = DeviceInformation(
                    str(data["serial"]),
                    str(data["serial"]),
                    data["device"],
                    data["susyid"],
                    "SMA",
                    data.get("sw_version", ""),
                )
                self._device_list[str(data["serial"])] = di
        return self._device_list

    # @override
    async def read(self, sensors: Sensors, deviceID: str | None = None) -> bool:
        """Read a set of keys.

        Args:
            sensors (Sensors): Sensors object containing Sensor objects to read

        Returns:
            bool: reading was successful
        """
        notfound = []
        data = await self._get_next_values(deviceID)
        for sensor in sensors:
            if sensor.key in data:
                value = data[sensor.key]
                if sensor.factor:
                    value /= sensor.factor
                sensor.value = value
            else:
                notfound.append(sensor.key)

        if notfound:
            _LOGGER.info(
                "No values for sensors: %s",
                ",".join(notfound),
            )

        return True

    # @override
    async def close_session(self) -> None:
        """Closes the session"""
        if self._transport:
            self._transport.close()
        self._sock.close()

    # @override
    async def detect(self, ip: str) -> List[DiscoveryInformation]:
        """Try to detect SMA devices"""
        discovered = []
        try:
            await self.new_session()
            start = time.time()
            while time.time() - start < 2.1:
                try:
                    data = await self._get_next_values()
                except TimeoutError:
                    continue
                if data["ip"].startswith(ip + ":"):
                    device = f'{data["device"]} ({data["serial"]})'
                    di = DiscoveryInformation(
                        data["ip"], "found", "speedwireem", None, "", device
                    )
                    if di not in discovered:
                        discovered.append(di)
        except (TimeoutError, SmaConnectionException):
            pass
        if len(discovered) == 0:
            discovered.append(
                DiscoveryInformation(
                    ip,
                    "failed",
                    "speedwireem",
                    None,
                    "no multicast packet received.",
                    "",
                )
            )
        return discovered

    # @override
    async def get_debug(self) -> Dict[str, Any]:
        """Return a dict with all debug information."""
        debug_info = {
            "last_packet": (
                base64.b64encode(self.di.last_packet).decode("ascii")
                if self.di.last_packet is not None
                else ""
            ),
            "last_valid_packet": (
                base64.b64encode(self.di.last_valid_packet).decode("ascii")
                if self.di.last_valid_packet is not None
                else ""
            ),
            "last_data": self.di.last_data,
            "serial": list(self.di.serial),
            "protocol": list(self.di.protocol),
        }
        return debug_info

    # @override
    def set_options(self, options: Dict[str, Any]) -> None:
        """Set options"""
        for key, item in options.items():
            if key.lower() == "bindingaddr":
                addrs = str(item).split(",")
                self._bindingAddr.extend(addrs)

    # @override
    async def set_parameter(
        self, sensor: Sensor, value: int, deviceId: str | None = None
    ) -> None:
        """Set Parameters."""

    def _getDiscoverySocket(self) -> socket.socket:
        multicast_group = "239.12.255.254"
        server_address = ("", 9522)

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(server_address)
        if len(self._bindingAddr) == 0:
            self._bindingAddr = [socket.INADDR_ANY]
            mreq = struct.pack(
                "4sL", socket.inet_aton(multicast_group), socket.INADDR_ANY
            )
            sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
        else:
            _LOGGER.info("Binding to %s" % self._bindingAddr)
            for addr in self._bindingAddr:
                try:
                    mreq = struct.pack(
                        "4s4s",
                        socket.inet_aton(multicast_group),
                        socket.inet_aton(addr),
                    )
                    sock.setsockopt(
                        socket.IPPROTO_IP,
                        socket.IP_ADD_MEMBERSHIP,
                        socket.inet_aton(multicast_group) + socket.inet_aton(addr),
                    )
                except BaseException as exc:
                    raise RuntimeError(
                        "Could not start multicast for %s. IP of the Interfaces must be used!"
                        % addr
                    ) from exc
        self._sock = sock
        return self._sock

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Called if connection is made"""
        self.transport = transport

    def error_received(self, exc: Exception) -> None:
        """Called by error."""
        _LOGGER.error("%s error occurred: %s", type(exc), exc)

    def connection_lost(self, exc: Exception) -> None:
        """Called by connection lost."""
        # _LOGGER.error("Socket closed, stop the event loop %s %s", type(exc), exc)

    def datagram_received(self, p: bytes, addr: tuple[str, int]) -> dict[str, Any]:
        """Decode a Speedwire-Packet

        Args:
            p: Network-Packet

        Returns:
            dict: Dict with all the decoded information
        """
        self.di.last_packet = p
        sw = speedwireHeader.from_packed(p[0:18])
        self.di.protocol.add(f"{sw.protokoll:04x}")
        if not sw.check6069():
            return {}
        sw6069 = speedwireHeader6069.from_packed(p[18:28])
        m = (addr[0], addr[1], sw6069.timestamp)
        if self.di.last_packet_metadata == m:
            return {}
        self.di.last_packet_metadata = m
        data: dict[str, Any] = {}
        data["protocolID"] = sw.protokoll
        data["susyid"] = sw6069.src_susyid
        data["device"] = SMATagList.get(data["susyid"], "unknown")
        data["serial"] = sw6069.src_serial
        data["ip"] = addr[0] + ":" + str(addr[1])
        length = sw.smanet2_length + 16
        pos = 28
        while pos < length:
            value: Any = None
            mchannel = int.from_bytes(p[pos : pos + 1], byteorder="big")
            mvalueindex = int.from_bytes(p[pos + 1 : pos + 2], byteorder="big")
            mtyp = int.from_bytes(p[pos + 2 : pos + 3], byteorder="big")
            mtariff = int.from_bytes(p[pos + 3 : pos + 4], byteorder="big")
            obis = f"{mvalueindex}:{mtyp}:{mtariff}"
            if mtyp in [4, 8]:
                # 4 actucal / current => 8 Bytes
                # 8 counter / sum => 12 Bytes
                value = int.from_bytes(p[pos + 4 : pos + 4 + mtyp], byteorder="big")
                pos += 4 + mtyp
            elif mchannel == 144 and mtyp == 0:
                value = f"{p[pos + 4]}.{p[pos + 5]}.{p[pos + 6]}.{chr(p[pos + 7])}"
                obis = "sw_version"
                pos += 4 + 4
            else:
                _LOGGER.debug(
                    "Unknown packet in speedwire: "
                    + str(mchannel)
                    + " "
                    + str(mvalueindex)
                    + " "
                    + str(mtyp)
                    + " "
                    + str(mtariff)
                )
                pos += 4 + 4
            data[obis] = value

        # Statistics & Co
        self.di.serial.add(data["serial"])
        self.di.last_valid_packet = p
        self.di.last_data = data
        if self._data_received is not None and not self._data_received.done():
            if self._expected_device is None or self._expected_device == str(
                data["serial"]
            ):
                self._data_received.set_result(data)
        return data
